from __future__ import absolute_import

import logging
import gzip
import os
import sys
import string
import random
import pysam
from io import TextIOWrapper
from multiprocessing import Process
from signal import signal, SIGPIPE, SIG_DFL
from bioat.logger import set_logging_level

# from bioat.lib.libdataclasses import Bam

signal(SIGPIPE, SIG_DFL)


class BamTools:
    """Bam toolbox."""

    def __init__(self):
        pass

    def mpileup_to_table(
            self,
            mpileup: str,
            output: str = sys.stdout,
            threads: int = os.cpu_count() - 1,
            mutation_number_threshold: int = 0,
            temp_dir: str = f"/tmp/bioat_{''.join(random.sample(string.ascii_letters + string.digits, 16))}",
            remove_temp: bool = True,
            log_level: str = 'INFO'
    ):
        """Convert mpileup file to info file.

        :param mpileup: samtools mpileup format file
        :param output: Output parsed file
        :param threads: Multiple threads number
        :param mutation_number_threshold: Only contain mutation info go to the output, set 0 mean output all site
        :param temp_dir: Where to keep temp files, default is /tmp
        :param remove_temp: Where to keep temp files, default is /tmp
        :param log_level: 'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'
        """
        set_logging_level(level=log_level)
        # set logger
        lib_name = __name__
        function_name = sys._getframe().f_code.co_name
        logger = logging.getLogger(f'{lib_name}.{function_name} ==> ')

        def _temp_split_bam(mpileup, threads, temp_dir):
            # creat temp file name and open file
            logger.debug("Making temp files...")

            input_file_basename = os.path.basename(mpileup)

            temp_filename_list = []
            temp_file_list = []

            os.makedirs(temp_dir, exist_ok=True)

            for index in range(threads):
                # make filename
                temp_file_basename = "temp_" + input_file_basename + "." + str(index) + "." + "".join(
                    random.sample(string.ascii_letters + string.digits, 16)) + '.gz'
                temp_file_name = os.path.join(temp_dir, temp_file_basename)
                temp_filename_list.append(temp_file_name)

                # open file
                temp_file = gzip.open(temp_file_name, "wt")
                temp_file_list.append(temp_file)

            # counting input file line number
            logger.debug("Counting input file...")

            total_input_line_num = 0
            input_file = gzip.open(mpileup, "rt") if mpileup.endswith('.gz') else open(mpileup, "rt")

            # count total_input_line_num
            for _ in input_file:
                total_input_line_num += 1

            if total_input_line_num % threads == 0:
                each_file_line_num = (total_input_line_num // threads)
            else:
                each_file_line_num = (total_input_line_num // threads) + 1

            input_file.close()
            logger.debug("Done!")

            # split temp files
            # write into output filename
            input_file = gzip.open(mpileup, "rt") if mpileup.endswith('.gz') else open(mpileup, "rt")

            for index, line in enumerate(input_file):
                file_index = index // each_file_line_num
                temp_file_list[file_index].write(line)

            # close output filename
            input_file.close()
            [temp_file.close() for temp_file in temp_file_list]
            logger.debug("Make temp files done!")

            return temp_filename_list

        def _merge_out_bmat(temp_filename_list, output, remove_temp=True):
            if type(output) is str:
                output = gzip.open(output, "wt") if output.endswith('.gz') else open(output, "wt")
            # write header
            header = [
                "chr_name",
                "chr_index",
                "ref_base",
                "A",
                "G",
                "C",
                "T",
                "del_count",
                "insert_count",
                "ambiguous_count",
                "deletion",
                "insertion",
                "ambiguous",
                "mut_num"
            ]
            output.write("\t".join(header) + "\n")

            # open temp bmat file
            for temp_filename in temp_filename_list:
                # write file
                with gzip.open(temp_filename + ".bmat.gz", "rt") as temp_file:
                    for line in temp_file:
                        output.write(line)

            # remove temp files
            if remove_temp:
                for temp_filename in temp_filename_list:
                    os.remove(temp_filename)
                    os.remove(temp_filename + ".bmat.gz")
                os.rmdir(temp_dir)

        # make temp files
        temp_filename_list = _temp_split_bam(mpileup=mpileup, threads=threads, temp_dir=temp_dir)
        # multiple processing part
        logger.debug("Parsing files...")

        procs_list = []

        for index, temp_mpileup in enumerate(temp_filename_list):
            temp_output = temp_mpileup + ".bmat.gz"
            # add multiprocess
            sub_proc = Process(target=_run_mpileup_to_table,
                               args=(temp_mpileup, temp_output, mutation_number_threshold,))
            procs_list.append(sub_proc)
            sub_proc.start()

        for sub_proc in procs_list:
            sub_proc.join()
        logger.debug("Done!")
        # merge output files
        logger.debug("Merging files...")
        _merge_out_bmat(temp_filename_list=temp_filename_list, output=output, remove_temp=remove_temp)
        logger.debug("Done!...")

    def remove_clip(
            self,
            input: str = sys.stdin,
            output: str = sys.stdout,
            threads: int = 1,
            output_fmt: str = 'SAM',
            remove_as_paired: bool = True,
            max_clip: int = 0,
            log_level: str = 'INFO'
    ):
        """Remove softclip reads in BAM file.

        :param input: BAM file sorted by queryname with soft/hard clip reads, pipe stdin is supported

                        [samtools view -h foo_sort_name.bam | bioat bam remove_clip <flags>]
        :param output: BAM file sorted by queryname without soft/hard clip reads, pipe stdout is supported

                        [bioat bam remove_clip <flags> | wc -l]
                        [bioat bam remove_clip <flags> | samtools view ....]
        :param threads: threads used by pysam and samtools core
        :param output_format: BAM, SAM
        :param remove_as_paired: True, False. remove single clip(False) / clip and its pair read(True)
        :param max_clip: the maximum clips allowed per read
        :param log_level: 'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'
        """
        set_logging_level(level=log_level)
        # set logger
        lib_name = __name__
        function_name = sys._getframe().f_code.co_name
        logger = logging.getLogger(f'{lib_name}.{function_name} ==> ')

        save = pysam.set_verbosity(0)  # https://github.com/pysam-developers/pysam/issues/939

        if isinstance(input, str):  # not stdin
            bam_in = pysam.AlignmentFile(input, "r", threads=threads, check_sq=False)
        elif isinstance(input, TextIOWrapper):  # stdin
            bam_in = pysam.AlignmentFile("-", "r", threads=threads, check_sq=False)
        else:
            exit(1)
        pysam.set_verbosity(save)

        try:
            so = bam_in.header['HD']['SO']
        except KeyError:
            so = None

        if so:
            if so != "queryname":
                logger.fatal(f"the input BAM|SAM must be sorted by name and has header [SO:queryname]!\n"
                             f"your header: [SO:{so}]\n")
                exit(1)
        else:
            logger.warning(f"the input BAM|SAM must be sorted by name and has header [SO:queryname]!\n"
                           "your bam file does not have a header\n")

        write_mode = 'wb' if output_fmt.upper() == "BAM" else 'w'
        bam_out = pysam.AlignmentFile(output, write_mode, template=bam_in)

        # Iterate through reads.
        read1, read2 = None, None

        for read in bam_in:
            if not read.is_paired or read.is_unmapped or read.mate_is_unmapped:
                # or read.is_duplicate:
                # only use mapped paired reads
                continue

            # ------------------------------------------------------------------->>>>>>>>>>
            # 如果先看到read1，将read1赋给read， 判断是否read1、read2都存在
            #    不都存在：
            #        即read2不存在，进入下个循环
            #    都存在，判断query name是否一致
            # ------------------------------------------------------------------->>>>>>>>>>
            if read.is_read2:
                # if first is read2, then use it to find read2
                read2 = read
            else:
                # if first is read1, then use it to find read1, and next query
                read1 = read
                read2 = None
                continue

            # if read1 not None and read2 not None pass to below
            if read1 and read2 and read1.query_name == read2.query_name:
                # print("found a pair!")
                # print(f'Cigar: Read1-{read1.cigarstring}, Read2-{read1.cigarstring}')
                # print(read.get_cigar_stats())
                softclip1 = read1.get_cigar_stats()[0][4]
                softclip2 = read2.get_cigar_stats()[0][4]

                if not remove_as_paired:
                    if softclip1 <= max_clip:
                        bam_out.write(read1)
                    if softclip2 <= max_clip:
                        bam_out.write(read2)
                else:
                    if softclip1 <= max_clip and softclip2 <= max_clip:
                        bam_out.write(read1)
                        bam_out.write(read2)
            else:
                logger.fatal(
                    'something wrong with read1/2 pairs, they may have different read id or there is None in them')
                logger.fatal(f'read1={read1}, read2={read2}')
                logger.fatal(f'read1_id={read1.query_name}, read2_id={read2.query_name}')
                exit(1)

        bam_in.close()
        bam_out.close()


def _run_mpileup_to_table(temp_mpileup, temp_output, mutation_number_threshold):
    # open input file
    temp_mpileup = gzip.open(temp_mpileup, "rt") if temp_mpileup.endswith('.gz') else open(temp_mpileup, "rt")
    # open output file
    temp_output = gzip.open(temp_output, "wt") if temp_output.endswith('.gz') else open(temp_output, "wt")

    # parse mpileup file
    for line in temp_mpileup:
        data = line.strip().split('\t')
        chr_name = data[0]
        bp = data[1]
        bases = data[4].upper()
        ref = data[2].upper()
        types = {'A': 0, 'G': 0, 'C': 0, 'T': 0, '-': [], '+': [], 'Not': []}

        # coverage > 0
        i = 0
        while i < len(bases):
            base = bases[i]

            if base == '^':
                i += 2

            elif base == '$':
                i += 1

            elif base == '-':
                i += 1
                del_str_num = ""

                while bases[i].isdigit():
                    del_str_num += bases[i]
                    i += 1

                delNum = int(del_str_num)
                delSeq = ""
                for a in range(delNum):
                    delSeq += bases[i]
                    i += 1
                types['-'].append(delSeq)

            elif base == '*':
                types['-'].append(bases[i])
                i += 1

            elif base == '+':
                i += 1
                add_str_num = ""

                while bases[i].isdigit():
                    add_str_num += bases[i]
                    i += 1

                addNum = int(add_str_num)
                addSeq = ""
                for a in range(addNum):
                    addSeq += bases[i]
                    i += 1
                types['+'].append(addSeq)

            elif (base == '.') or (base == ','):
                types[ref] += 1
                i += 1

            else:
                if base in types:
                    types[base] += 1
                else:
                    types['Not'].append(base)
                i += 1

        adds = '.'
        adds_count = 0
        if len(types['+']) > 0:
            adds = ','.join(types['+'])
            adds_count = len(types['+'])

        dels = "."
        dels_count = 0
        if len(types['-']) > 0:
            dels = ",".join(types['-'])
            dels_count = len(types['-'])

        amb = '.'
        amb_count = 0
        if len(types['Not']) > 0:
            amb = ','.join(types['Not'])
            amb_count = len(types['Not'])

        # get other mutation number
        if ref in types:
            mut_num = types["A"] + types["T"] + types["G"] + types["C"] - types[ref]
        else:
            mut_num = 0

        out_list = [
            chr_name,
            bp,
            ref,
            types['A'],
            types['G'],
            types['C'],
            types['T'],
            dels_count,
            adds_count,
            amb_count,
            dels,
            adds,
            amb,
            mut_num
        ]

        # make a filter
        if mutation_number_threshold:
            if mut_num >= mutation_number_threshold:
                temp_output.write("\t".join(map(str, out_list)) + "\n")
        else:
            temp_output.write("\t".join(map(str, out_list)) + "\n")

    temp_mpileup.close()
    temp_output.close()


if __name__ == '__main__':
    pass



from __future__ import absolute_import
# from bioat.api.internet_speed import main as speedtest


class SystemTools:
    def __init__(self):
        # self.test_internet_speed = speedtest()
        pass

#     def __convert_size(self, size, mode='b') -> float:
#         """将单位为Byte的数字转为其它单位
#
#         :param size: int, 单位为Byte的数字
#         :param mode: str, 单位，b、k、m、g、t或者B、K、M、G、T
#         :return:
#         """
#         assert mode in list('bkmgtBKMGT')
#         mode = mode.lower()
#
#         if mode == 'b':
#             return float(size)
#         elif mode == 'k':
#             return size / 1024.0
#         elif mode == 'm':
#             return size / 1024.0 / 1024.0
#         elif mode == 'g':
#             return size / 1024.0 / 1024.0 / 1024.0
#         elif mode == 't':
#             return size / 1024.0 / 1024.0 / 1024.0 / 1024.0
#         else:
#             return None
#
#     def __get_size(self, path, mode='b'):
#         """获取路径占用大小.
#         通过递归函数获取路径占用大小
#         :param path: str, 目标路径
#         :param mode: str, 单位，b、k、m、g、t或者B、K、M、G、T
#         :return: str, 占用大小
#         """
#         assert mode in list('bkmgtBKMGT')
#         mode = mode.lower()
#
#         # 目标是否存在
#         if os.path.exists(path):
#             # 目标路径是一个文件
#             if os.path.isfile(path):
#                 # 无序遍历，直接返回即可
#                 byte_size = os.path.getsize(path)
#                 return self.__convert_size(size=byte_size, mode=mode)
#
#             # 目标路径是一个文件夹
#             byte_size = 0.0
#             files = os.listdir(path)
#
#             for file in files:
#                 cur_path = os.path.join(path, file)
#                 try:
#                     byte_size += self.__get_size(cur_path)
#                 except TypeError:
#                     continue
#
#             return self.__convert_size(size=byte_size, mode=mode)
#         else:
#             # print(f'{path} not exists!')
#             pass
#
#
# if __name__ == '__main__':
#     # 大路径
#     aim_paths = [
#         # '/lustre1/chengqiyi_pkuhpc/zhaohn/2.disk_occupy/1T/fastq', # for test
#         # '/lustre1/chengqiyi_pkuhpc/zhaohn/2.disk_occupy/1T.1/1T/fastq', # for test\
#         #
#         '/lustre1/chengqiyi_pkuhpc/zhaohn',
#         '/lustre1/chengqiyi_pkuhpc/JiangZhe',
#         '/lustre1/chengqiyi_pkuhpc/LiuCong',
#         '/lustre1/chengqiyi_pkuhpc/liuwenqing',
#         '/lustre1/chengqiyi_pkuhpc/lubo',
#         '/lustre1/chengqiyi_pkuhpc/shx',
#         '/lustre2/chengqiyi_pkuhpc/bds',
#         '/lustre3/chengqiyi_pkuhpc/folder_for_learning/bds',
#         '/lustre2/chengqiyi_pkuhpc/dyy',
#         '/lustre2/chengqiyi_pkuhpc/LK',
#         '/lustre2/chengqiyi_pkuhpc/log',
#         '/lustre2/chengqiyi_pkuhpc/LSY',
#         '/lustre2/chengqiyi_pkuhpc/RenweiLi',
#         '/lustre2/chengqiyi_pkuhpc/wuhao',
#         '/lustre2/chengqiyi_pkuhpc/xiongxs_tmp',
#         '/lustre2/chengqiyi_pkuhpc/zhangxt',
#         '/lustre3/chengqiyi_pkuhpc/folder_for_learning/zhangxt',
#         '/lustre2/chengqiyi_pkuhpc/zml',
#         '/lustre3/chengqiyi_pkuhpc/Alumni/mshq',
#         '/lustre3/chengqiyi_pkuhpc/Alumni/Shuxiaoting',
#         '/lustre3/chengqiyi_pkuhpc/Alumni/SJH',
#         '/lustre3/chengqiyi_pkuhpc/Alumni/WK',
#         '/lustre3/chengqiyi_pkuhpc/folder_for_learning/dlt',
#         '/lustre3/chengqiyi_pkuhpc/folder_for_learning/Leizhixin',
#         '/lustre3/chengqiyi_pkuhpc/folder_for_learning/LiuJiangle',
#         '/lustre3/chengqiyi_pkuhpc/folder_for_learning/lixy',
#         '/lustre3/chengqiyi_pkuhpc/folder_for_learning/lzc',
#         '/lustre3/chengqiyi_pkuhpc/folder_for_learning/test2',
#         '/lustre3/chengqiyi_pkuhpc/folder_for_learning/ydy',
#         '/lustre3/chengqiyi_pkuhpc/folder_for_learning/ZhuangYuan',
#     ]
#
#     out_path = '/lustre1/chengqiyi_pkuhpc/zhaohn/__diskusage.all_user.csv'
#
#     dt_size = {key: [] for key in aim_paths}
#
#     for aim_path in aim_paths:
#         file_sizes = []
#         files = os.listdir(aim_path)
#
#         # print("aim_path: ", aim_path)
#         for file in files:
#             # continue
#             cur_path = os.path.join(aim_path, file)
#             # print("cur_path: ", cur_path)
#             if os.path.isdir(cur_path):
#                 try:
#                     file_size = (file, get_size(cur_path, 't'))
#                     file_sizes.append(file_size)
#                     print(f"INFO:    Processed {file_size[0]}  {file_size[1]:.3f}TB")
#                 except Exception as err:  # 捕捉所有异常
#                     print(f"WARNING: {err} @ {cur_path}, Skip ...")
#                 except FileNotFoundError:
#                     print(f"FileNotFoundError @ {cur_path}")
#                     continue
#                 except Exception as err:
#                     print(f"Exception @ {cur_path}: {err}")
#
#         # print("file_sizes: ", file_sizes)
#         file_sizes.sort(key=lambda t: t[1], reverse=True)
#         dt_size[aim_path] = file_sizes
#     date_time = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime())
#
#     with open(out_path, 'wt') as f:
#         f.write(f'{date_time},by,Huanan Zhao,\n,,,\n')
#         f.write("USER,PATH,TOTAL_SIZE,UNIT\n")
#         # print(dt_size)
#
#         for key in dt_size.keys():
#             user = os.path.basename(key)
#             size = sum([item[1] for item in dt_size[key]])
#             f.write(f'{user},{key},{size:>3f},TB\n')
#
#     print(f'one loop done! @ {date_time}')



import dataclasses


@dataclasses.dataclass
class Bed:
    chromosome: str
    start: int
    end: int
    name: str
    score: int
    strand: str


@dataclasses.dataclass
class Bam:
    pass


@dataclasses.dataclass
class Fasta:
    header: str
    sequence: str

    def format_fasta(self, length: 80):
        # 格式化输出fasta的sequence
        pass


@dataclasses.dataclass
class Fastq:
    # https://upload-images.jianshu.io/upload_images/7976641-18ceaafbce3d93d7.png?imageMogr2/auto-orient/strip|imageView2/2/w/1199/format/webp
    header: str
    sequence: str
    extras: str
    quality: str

    def get_phread_info(self):
        # 判断是Phred33还是Phred64
        # https://www.jianshu.com/p/248308513e2e
        pass


@dataclasses.dataclass
class VCF:
    pass


from __future__ import absolute_import
import pandas as pd
import logging
import sys
from bioat.logger import set_logging_level

__all__ = ['set_option']


def set_option(max_colwidth: int = 40, display_width: int = 120, display_max_columns: int = None,
               display_max_rows: int = 50):
    # set logger
    lib_name = __name__
    function_name = sys._getframe().f_code.co_name
    logger = logging.getLogger(f'{lib_name}.{function_name} ==> ')

    logger.info(f'set pandas.set_options: max_colwidth={max_colwidth}')
    pd.set_option("max_colwidth", max_colwidth)  # column最大宽度
    logger.info(f'set pandas.set_options: display.width={display_width}')
    pd.set_option("display.width", display_width)  # dataframe宽度
    logger.info(f'set pandas.set_options: display.max_columns={display_max_columns}')
    pd.set_option("display.max_columns", display_max_columns)  # column最大显示数
    logger.info(f'set pandas.set_options: display.max_rows={display_max_rows}')
    pd.set_option("display.max_rows", display_max_rows)  # row最大显示数
    return None



# from __future__ import absolute_import
# import os
# import sys
# import logging
# from bioat.logger import set_logging_level
# from bioat.lib.libdetect_seq import (
#     check_input_bam_file,
#     clear_temp_files_by_dict,
#     split_mpmat_by_chr_name,
#     make_qvalue_with_BH_method,
#     multi_thread_chr_bam_mut_count,
#     calculate_effective_genome,
#     count_dict_sum_up,
#     back_all_normalization_scale_dict,
#     multi_run_mpmat_poisson_test,
#     merge_split_files,
#     _log_cmd_str,
#     _check_file_exist
# )
#
#
# class DetectSeq:
#     def __init__(self):
#         pass
#
#     def find_significant_mpmat(
#             self,
#             mpmat_table: str,
#             ctrl_BAM: str,
#             treat_BAM: str,
#             reference: str,
#             output: str = "./poisson_output.tsv",
#             thread: int = 1,
#             query_mutation_type: str = "CT,GA",
#             keep_temp_file=False,
#             mpmat_filter_info_col_index: int = -1,
#             mpmat_block_info_col_index: int = -1,
#             region_block_mut_num_cutoff: int = 2,
#             query_mut_min_cutoff: int = 1,
#             query_mut_max_cutoff: int = 16,
#             total_mut_max_cutoff: int = 20,
#             other_mut_max_cutoff: int = 22,
#             seq_reads_length: int = 150,
#             scale_reads_count: int = 1e6,
#             lambda_method: str = "ctrl_max",
#             poisson_method: str = "mutation",
#             log_level: str = 'INFO'
#     ):
#         """
#         A tool to find significant enrichment mpmat region. You can set a lot of parameters with this program,
#         usually default parameters can work well.
#
#         :param mpmat_table: .mpmat table, which can be generated from <pmat-merge.py> code
#         :param ctrl_BAM: Control BAM file
#         :param treat_BAM: Treatment BAM file
#         :param reference: Genome FASTA file
#         :param output: Output Poisson test result, default=poisson_output.tsv
#
#         :param thread: Number of threads used to process data
#         :param query_mutation_type: Query mutation type, which will be considered as mutation signals, default=CT,GA
#         :param keep_temp_file: If keep temp files, default=False
#         :param mpmat_filter_info_col_index: Column index for filter info, -1 means no such info. Default=-1. Usually can set as 13
#         :param mpmat_block_info_col_index: Column index for region block info, -1 means no such info. Default=-1.
#         :param region_block_mut_num_cutoff: Site filter cutoff, if a site has a mutation signal >= this cutoff in ctrl sample, the site will be blocked in the downstream analysis. Default=2
#         :param query_mut_min_cutoff: An alignment contains query mutation count lower than this cutoff in mpmat region will considered as 'non_mut_align', default=1
#         :param query_mut_max_cutoff: An alignment contain mutation number higher than this, considered as 'query_high_mismatch', which often caused by Bismark mapping error, default=16
#         :param total_mut_max_cutoff: An alignment contain mutation number higher than this, considered as 'total_high_mismatch', default=20
#         :param other_mut_max_cutoff: An alignment contain mutation number higher than this, considered as 'other_high_mismatch', default=12
#         :param seq_reads_length: Sequencing reads length, default=150
#         :param scale_reads_count: Scaled final output region signal, default=1000000 Usually, default=1e6, means CPM
#         :param lambda_method: Can be set as 'ctrl_max', 'treat_max', 'max', 'raw', default=ctrl_max
#         :param poisson_method: Can be set as 'mutation' OR 'all', default=mutation. 'mutation' means only use mutation alignments to run Poisson test,'all' means use all alignments to run Poisson, which similar to MACS2
#         :param log_level: 'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'
#         :return:
#         """
#         set_logging_level(level=log_level)
#         # set logger
#         lib_name = __name__
#         function_name = sys._getframe().f_code.co_name
#         logger = logging.getLogger(f'{lib_name}.{function_name} ==> ')
#
#         ARGS = dict(
#             mpmat_table=mpmat_table,
#             output=output,
#             ctrl_BAM=ctrl_BAM,
#             treat_BAM=treat_BAM,
#             reference=reference,
#             thread=thread,
#             query_mutation_type=query_mutation_type,
#             verbose=log_level,
#             keep_temp_file=keep_temp_file,
#             mpmat_filter_info_col_index=mpmat_filter_info_col_index,
#             mpmat_block_info_col_index=mpmat_block_info_col_index,
#             region_block_mut_num_cutoff=region_block_mut_num_cutoff,
#             query_mut_min_cutoff=query_mut_min_cutoff,
#             query_mut_max_cutoff=query_mut_max_cutoff,
#             total_mut_max_cutoff=total_mut_max_cutoff,
#             other_mut_max_cutoff=other_mut_max_cutoff,
#             seq_reads_length=seq_reads_length,
#             scale_reads_count=scale_reads_count,
#             lambda_method=lambda_method,
#             poisson_method=poisson_method
#         )
#
#         # output full cmd
#         sys.stderr.write("\n" + "-" * 80 + "\n")
#         sys.stderr.write(_log_cmd_str(args=ARGS))
#
#         # check file exist
#         sys.stderr.write("\n" + "-" * 80 + "\n")
#         check_res = _check_file_exist(args=ARGS)
#         if not check_res[0]:
#             raise IOError("Please make sure each file is exist!")
#
#         # check output file dir
#         if output == "stdout":
#             temp_dir = os.getcwd()
#         else:
#             temp_dir = os.path.dirname(output)
#
#         # fix ARGS
#         query_mutation_type = query_mutation_type.split(",")
#
#         # ---------------------------------------------------------------------------->>>>>>>>>>
#         # Part I split mpmat
#         # ---------------------------------------------------------------------------->>>>>>>>>>
#         # split
#         sys.stderr.write("\n" + "-" * 80 + "\n")
#         split_mpmat_filename_dict = split_mpmat_by_chr_name(
#             input_mpmat_filename=mpmat_table,
#             temp_dir=temp_dir,
#             force_temp_dir=True)
#
#         # meta dict
#         meta_dict = {"filename": {}}
#         meta_dict["filename"]["mpmat_region_AddBlockInfo"] = mpmat_table
#         meta_dict["filename"]["mpmat_region_AddBlockInfo_split"] = split_mpmat_filename_dict.copy()
#
#         # make chr order
#         ref_order_dict = {}
#         ref_order_list = []
#
#         for index, chr_name in enumerate(meta_dict["filename"]["mpmat_region_AddBlockInfo_split"]["chr_name_order"]):
#             ref_order_dict[chr_name] = index
#             ref_order_list.append(chr_name)
#
#         # ---------------------------------------------------------------------------->>>>>>>>>>
#         # Part II calculation of normalization scale factor
#         # ---------------------------------------------------------------------------->>>>>>>>>>
#         sys.stderr.write("-" * 80 + "\n")
#         logger.info("Calculating ctrl sample Poisson background...")
#
#         ctrl_poisson_bg_meta_dict = multi_thread_chr_bam_mut_count(
#             input_bam_filename=ARGS.ctrl_BAM,
#             select_chr_name_list=ref_order_list,
#             ref_genome_filename=ARGS.reference,
#             thread_num=ARGS.thread,
#             query_mut_type_list=query_mutation_type,
#             query_mut_min_cutoff=ARGS.query_mut_min_cutoff,
#             query_mut_max_cutoff=ARGS.query_mut_max_cutoff,
#             total_mut_max_cutoff=ARGS.total_mut_max_cutoff,
#             other_mut_max_cutoff=ARGS.other_mut_max_cutoff,
#             log_verbose=ARGS.verbose)
#
#         sys.stderr.write("-" * 80 + "\n")
#         logger.info("Calculating treat sample Poisson background...")
#
#         treat_poisson_bg_meta_dict = multi_thread_chr_bam_mut_count(
#             input_bam_filename=ARGS.treat_BAM,
#             select_chr_name_list=ref_order_list,
#             ref_genome_filename=ARGS.reference,
#             thread_num=ARGS.thread,
#             query_mut_type_list=query_mutation_type,
#             query_mut_min_cutoff=ARGS.query_mut_min_cutoff,
#             query_mut_max_cutoff=ARGS.query_mut_max_cutoff,
#             total_mut_max_cutoff=ARGS.total_mut_max_cutoff,
#             other_mut_max_cutoff=ARGS.other_mut_max_cutoff,
#             log_verbose=ARGS.verbose)
#
#         # genome effective length
#         sys.stderr.write("-" * 80 + "\n")
#         genome_base_count_dict = calculate_effective_genome(ref_genome_filename=ARGS.reference,
#                                                             log_verbose=ARGS.verbose)
#
#         # make meta data dict
#         meta_data_dict = {"ctrl": count_dict_sum_up(ctrl_poisson_bg_meta_dict["count_dict"]),
#                           "treat": count_dict_sum_up(treat_poisson_bg_meta_dict["count_dict"])}
#
#         # make scale dict
#         scale_factor_dict, normalize_scale_factor_dict, genome_bg_dict = back_all_normalization_scale_dict(
#             meta_data_dict=meta_data_dict,
#             genome_base_count_dict=genome_base_count_dict,
#             seq_reads_length=ARGS.seq_reads_length,
#             norm_scale_reads_count=int(ARGS.scale_reads_count))
#
#         # ---------------------------------------------------------------------------->>>>>>>>>>
#         # Part III Poisson test on mpmat region
#         # ---------------------------------------------------------------------------->>>>>>>>>>
#         poisson_out_split_dict = multi_run_mpmat_poisson_test(
#             mpmat_block_split_dict=meta_dict["filename"]["mpmat_region_AddBlockInfo_split"],
#             ctrl_bam_filename=ARGS.ctrl_BAM,
#             treat_bam_filename=ARGS.treat_BAM,
#             ref_genome_fa_filename=ARGS.reference,
#             scale_factor_dict=scale_factor_dict,
#             normalize_scale_factor_dict=normalize_scale_factor_dict,
#             genome_bg_dict=genome_bg_dict,
#             lambda_bg_method=ARGS.lambda_method,
#             poisson_method=ARGS.poisson_method,
#             log_verbose=ARGS.verbose,
#             thread=ARGS.thread,
#             region_block_mut_num_cutoff=ARGS.region_block_mut_num_cutoff,
#             reads_query_mut_min_cutoff=ARGS.query_mut_min_cutoff,
#             reads_query_mut_max_cutoff=ARGS.query_mut_max_cutoff,
#             reads_total_mut_max_cutoff=ARGS.total_mut_max_cutoff,
#             reads_other_mut_max_cutoff=ARGS.other_mut_max_cutoff,
#             mpmat_filter_info_col_index=ARGS.mpmat_filter_info_col_index,
#             mpmat_block_info_col_index=ARGS.mpmat_block_info_col_index
#         )
#
#         meta_dict["filename"]["Poisson_out_split"] = poisson_out_split_dict.copy()
#         meta_dict["filename"]["Poisson_out_merge_NoFDR"] = ARGS.output + ".NoFDR"
#
#         # ---------------------------------------------------------------------------->>>>>>>>>>
#         # Part VI output Poisson test result
#         # ---------------------------------------------------------------------------->>>>>>>>>>
#         sys.stderr.write("-" * 80 + "\n")
#         merge_state, pval_list_str = merge_split_files(split_file_dict=meta_dict["filename"]["Poisson_out_split"],
#                                                        key_order_list=ref_order_list,
#                                                        out_filename=ARGS.output + ".NoFDR",
#                                                        header_list=None,
#                                                        in_sep="\t", out_sep="\t",
#                                                        log_verbose=ARGS.verbose,
#                                                        return_col_index=-1)
#
#         sys.stderr.write("-" * 80 + "\n")
#         logger.info("Calculating FDR value and make final output table...")
#
#         # change pval into float
#         pval_list = []
#         for pval_str in pval_list_str:
#             if pval_str != "NA":
#                 pval_list.append(eval(pval_str))
#             else:
#                 pval_list.append("NA")
#
#         # make qval with BH method
#         qval_list = make_qvalue_with_BH_method(pval_list)
#
#         # make final output
#         final_out_file = open(ARGS.output, "wt")
#         final_out_file.write("\t".join(header_list) + "\n")
#
#         with open(meta_dict["filename"]["Poisson_out_merge_NoFDR"], "rt") as no_fdr_file:
#             for index, line in enumerate(no_fdr_file):
#                 line_list = line.strip().split("\t")
#                 line_list.append(str(qval_list[index]))
#                 final_out_file.write("\t".join(line_list) + "\n")
#
#         final_out_file.close()
#
#         logger.info("Done!")
#
#         # ---------------------------------------------------------------------------->>>>>>>>>>
#         # Part VII remove temp files
#         # ---------------------------------------------------------------------------->>>>>>>>>>
#         if ARGS.keep_temp_file == "False":
#             sys.stderr.write("-" * 80 + "\n")
#             logger.info("Set <keep_temp_file> as False... Removing temp files...")
#
#             sys.stderr.write("-" * 80 + "\n")
#             logger.info("Removing mpmat split files...")
#             clear_temp_files_by_dict(temp_file_dict=meta_dict["filename"]["mpmat_region_AddBlockInfo_split"],
#                                      log_verbose=ARGS.verbose)
#
#             sys.stderr.write("-" * 80 + "\n")
#             logger.info("Removing Poisson test split files...")
#             clear_temp_files_by_dict(temp_file_dict=meta_dict["filename"]["Poisson_out_split"],
#                                      log_verbose=ARGS.verbose)
#
#             sys.stderr.write("-" * 80 + "\n")
#             logger.info("Removing Poisson test raw file...")
#             try:
#                 os.remove(meta_dict["filename"]["Poisson_out_merge_NoFDR"])
#             except Warning as w:
#                 logger.warning("Removing file error: \n\t%s" % meta_dict["filename"]["Poisson_out_merge_NoFDR"])
#
#         logger.info("Everything done!")
#
#
# # ---------------------------------------------------------------------------->>>>>>>>>>
# #  main part
# # ---------------------------------------------------------------------------->>>>>>>>>>
# if __name__ == '__main__':
#     main()



from __future__ import absolute_import
import bioat
import datetime
import os

project_toml: str = os.path.join(bioat.__path__[0], 'version.py')
sec = os.path.getmtime(project_toml)
__upgrade_date__ = datetime.date.fromtimestamp(sec)
__version__ = "0.2.9"

__author__ = 'Hua-nan ZHAO @ Tsinghua University'
__email__ = 'hermanzhaozzzz@gmail.com'
__doc_format__ = "restructuredtext"
__doc_address__ = "https://github.com/hermanzhaozzzz/bioat/tree/master/docs"
__repo_address__ = "https://github.com/hermanzhaozzzz/bioat"